///////////////////////////////////////////////////
// LED.cpp
// 720020-1175x
//
// Applied Data Systems
//
// See LED.h for a description and history
// 
///////////////////////////////////////////////////
#include <windows.h>
#include "led.h"

/////////////////////////////////////////////////////////////////////////////
// Defines
//
#define ON	1
#define OFF 0
#define	MEMORY_SIZE					0x10000  // 64kiB memory space; max allocation for VirtualAlloc is 64MiB
                                     // must be 64kiB aligned (VirtualAlloc() requirement)

/////////////////////////////////////////////////////////////////////////////
// VirtualCopy definition
//
extern "C" BOOL VirtualCopy(LPVOID lpvDestMem, LPVOID lpvSrcMem, 
							DWORD dwSizeInBytes, DWORD dwProtectFlag);

/////////////////////////////////////////////////////////////////////////////
// Global variables
//
static BYTE					g_state;
static CRITICAL_SECTION		g_cs;
static LPVOID				g_ledAddr;
static int					g_refCount;
static DWORD				*g_hLED;
static HANDLE				g_hThread;

/////////////////////////////////////////////////////////////////////////////
// Forward declarations
//
LPVOID MapAddress(DWORD baseAddress);
void UnmapAddress(LPVOID address);
void SetLED(DWORD *hLED, int value);
DWORD WINAPI HeartbeatThreadProc(LPVOID lpParameter);

/////////////////////////////////////////////////////////////////////////////
// DLL Entry Point
//
extern "C"
BOOL WINAPI DllMain(HANDLE hInstance, DWORD dwReason, LPVOID /*lpReserved*/)
{
	DEBUGMSG(TRUE, (_T("+DllMain\r\n")));

	switch (dwReason) 
	{

        case DLL_PROCESS_ATTACH :
			DEBUGMSG(TRUE, (_T("  DllMain : DLL_PROCESS_ATTACH\r\n")));
            break;

        case DLL_PROCESS_DETACH :
			DEBUGMSG(TRUE, (_T("  DllMain : DLL_PROCESS_ATTACH\r\n")));
            break;
            
        case DLL_THREAD_DETACH :
			DEBUGMSG(TRUE, (_T("  DllMain : DLL_PROCESS_ATTACH\r\n")));
            break;
            
        case DLL_THREAD_ATTACH :
			DEBUGMSG(TRUE, (_T("  DllMain : DLL_THREAD_ATTACH\r\n")));
            break;
            
        default :
			DEBUGMSG(TRUE, (_T("  DllMain : default\r\n")));
            break;
	}

	DEBUGMSG(TRUE, (_T("-DllMain\r\n")));

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// Init
// 
// Called by the Device Manager to initialize a device.
// Returns the hDeviceContext
//
extern "C"
__declspec(dllexport)
DWORD LED_Init(LPCTSTR pContext, 
			   LPCVOID lpvBusContext)
{
	DEBUGMSG(TRUE, (_T("+LED_Init\r\n")));
	
	// initialize the global critical section
	InitializeCriticalSection(&g_cs);	

	// initialize global variables
	g_state		= FALSE;
	g_ledAddr	= NULL;
	g_refCount	= 0;
	g_hLED		= NULL;
	g_hThread	= NULL;

	// initialize the LED to a known state (OFF)
	LPVOID addr = MapAddress(LED_ADDR_BASE);
	g_hLED = (DWORD*)addr;
	g_hLED += (LED_ADDR_OFFSET / sizeof(DWORD));

	SetLED(g_hLED, OFF);

	UnmapAddress(g_hLED);
	g_hLED = NULL;

	DEBUGMSG(TRUE, (_T("-LED_Init\r\n")));

	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Open
// 
// Opens a device for reading, writing, or both. An application indirectly 
// invokes this function when it calls CreateFile to open special device file names.
//
extern "C"
__declspec(dllexport)
DWORD LED_Open(DWORD hDeviceContext, 
			   DWORD AccessCode, 
			   DWORD ShareMode)
{
	DEBUGMSG(TRUE, (_T("+LED_Open\r\n")));

	if(g_refCount == 0)
	{
		// virtual alloc and get pointer to register for LED
		g_ledAddr = MapAddress(LED_ADDR_BASE);
		g_hLED = (DWORD*)g_ledAddr;
		g_hLED += (LED_ADDR_OFFSET / sizeof(DWORD));

		// increment a reference counter
		g_refCount++;
		DEBUGMSG(TRUE, (_T("  g_refCount = %i\r\n"), g_refCount));
	}

	DEBUGMSG(TRUE, (_T("  g_hLED = 0x%08x\r\n"), g_hLED));
	DEBUGMSG(TRUE, (_T("-LED_Open\r\n")));

	// multiple consumers can use this, so just return the refcount
	return g_refCount;
}

/////////////////////////////////////////////////////////////////////////////
// IOControl
// 
// Sends a command to a device.
//
extern "C"
__declspec(dllexport)
BOOL LED_IOControl(DWORD hOpenContext,
				   DWORD dwCode,
				   PBYTE pBufIn,
				   DWORD dwLenIn,
				   PBYTE pBufOut,
				   DWORD dwLenOut,
				   PDWORD pdwActualOut)
{
	DEBUGMSG(TRUE, (_T("+LED_IOControl\r\n")));

	BOOL ret	= FALSE;
	UINT i		= 0;

	dwLenOut = 0;
	pBufOut= NULL;

	switch(dwCode)
	{
		case IOCTL_LED_START_HEARTBEAT:
			DEBUGMSG(TRUE, (_T("  IOCTL_LED_START_HEARTBEAT\r\n")));
			// only start it if it's not running
			if(g_hThread == NULL)
			{
				g_hThread = CreateThread(NULL, 0, HeartbeatThreadProc, NULL, 0, NULL);
			}

			break;
		case IOCTL_LED_STOP_HEARTBEAT:
			DEBUGMSG(TRUE, (_T("  IOCTL_LED_STOP_HEARTBEAT\r\n")));
			// only stop it if it running
			if(g_hThread != NULL)
			{
				// let the thread know it should stop
				HANDLE h;
				h = CreateEvent(NULL, FALSE, TRUE, _T("IOCTL_LED_STOP_HEARTBEAT"));
				SetEvent(h);

				// NULL the thread handle
				g_hThread = NULL;
			}

			break;
		case IOCTL_LED_BLINK_NUMBER:

			DEBUGMSG(TRUE, (_T("  IOCTL_LED_BLINK_NUMBER\r\n")));
			
			BLINK_NUMBER_DATA *bnd;
			char numberChars[9];

			// make sure it's the right amount of data
			if(dwLenIn != sizeof(BLINK_NUMBER_DATA))
			{
				SetLastError(ERROR_INVALID_PARAMETER);
				ret = false;
				break;
			}

			// make the incoming data more friendly
			bnd = (BLINK_NUMBER_DATA*)pBufIn;

			// convert the number to a character array
			_itoa(bnd->Number, numberChars, 10);

			DEBUGMSG(TRUE, (_T("  Number: %i\r\n"), bnd->Number));

			// prevent multiple-call interference
			EnterCriticalSection(&g_cs);

			// start with LED off
			SetLED(g_hLED, 0);
			Sleep(100);
			
			UINT c;

			DEBUGMSG(TRUE, (_T("total chars: %i\r\n"), strlen(numberChars)));
			for(c = 0 ; c < strlen(numberChars) ; c++)
			{
				DEBUGMSG(TRUE, (_T("Char[%i] = %c (%i)"), i, numberChars[c], numberChars[c]));

				switch(numberChars[c])
				{
					case '0':
						// we can't blink "zero" times, so an LED "twiddle" represents zero
						for(i = 0 ; i < 3 ; i++)
						{
							SetLED(g_hLED, bnd->LEDIndex);
							Sleep(150);
							SetLED(g_hLED, 0);
							Sleep(150);
						}
												
						break;
					default:
						// starting with 0x31, blink once for the digits
						for(i = 0 ; i < (UINT)(numberChars[c] - 0x30) ; i++)
						{
							SetLED(g_hLED, bnd->LEDIndex);
							Sleep(300);
							SetLED(g_hLED, 0);
							Sleep(300);
						}
						break;
				} // switch(numberChars[i])
				
				// end of digit
				Sleep(bnd->InterDigitDelay);

			}// for(i = 0 ; i < strlen(numberChars) ; i++)

			DEBUGMSG(TRUE, (_T("\r\n")));

			// end of number
			Sleep(bnd->TerminateDelay - bnd->InterDigitDelay);

			// release the critical section
			LeaveCriticalSection(&g_cs);

			break;
		default:
			SetLastError(ERROR_INVALID_PARAMETER);
	}

	DEBUGMSG(TRUE, (_T("-LED_IOControl\r\n")));
	return ret;
}

/////////////////////////////////////////////////////////////////////////////
// Seek
// 
// Moves the data pointer in the device.
//
extern "C"
__declspec(dllexport)
DWORD LED_Seek(DWORD hOpenContext,
			   long Amount,
			   WORD Type)
{
	DEBUGMSG(TRUE, (_T("+LED_Seek\r\n")));

	// not implemented so do nothing
	SetLastError(ERROR_CALL_NOT_IMPLEMENTED);

	DEBUGMSG(TRUE, (_T("-LED_Seek\r\n")));

	return -1;
}

/////////////////////////////////////////////////////////////////////////////
// Read
// 
// Reads data from the device identified by the open context.
//
extern "C"
__declspec(dllexport)
DWORD LED_Read(DWORD hOpenContext,
			   LPVOID pBuffer,
			   DWORD Count)
{
	DEBUGMSG(TRUE, (_T("+LED_Read\r\n")));

	// get a usable pointer
	BYTE *p = (BYTE*)pBuffer;

	// simply return the value of the global state variable
	p[0] = g_state;
	Count = 1;

	DEBUGMSG(TRUE, (_T("-LED_Read\r\n")));

	// return the "bytes" read
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Write
// 
// Writes data to the device.
//
extern "C"
__declspec(dllexport)
DWORD LED_Write(DWORD hOpenContext,
				LPCVOID pBuffer,
				DWORD Count)
{
	DEBUGMSG(TRUE, (_T("+LED_Write\r\n")));

	// make sure that we're supposed to write a byte
	if(Count == 0)
		return 0;

	// get a usable pointer
	BYTE *p = (BYTE*)pBuffer;

	// make sure only one thread can set the state at a time
	EnterCriticalSection(&g_cs);

	// set the global state variable
	if(p[0] == 0)
	{
		g_state = OFF;
	}
	else
	{
		g_state = ON;
	}

	DEBUGMSG(TRUE, (_T("  setting state = %i\r\n"), g_state));

	// set the actual LED state
	SetLED(g_hLED, g_state);

	// release the synchronization object
	LeaveCriticalSection(&g_cs);

	DEBUGMSG(TRUE, (_T("-LED_Write\r\n")));

	// we really only write a bit, but we can't return 1/8
	return 1;
}

/////////////////////////////////////////////////////////////////////////////
// Close
// 
// Closes the device context identified by hOpenContext.
//
extern "C"
__declspec(dllexport)
BOOL LED_Close(DWORD hOpenContext)
{
	DEBUGMSG(TRUE, (_T("+LED_Close\r\n")));

	// decrease the reference count
	g_refCount--;

	// if we're at zero, release the memory
	if(g_refCount == 0)
	{
		UnmapAddress(g_ledAddr);
		g_hLED = NULL;
	}

	if(g_hThread != NULL)
	{
		// let the thread know it should stop
		HANDLE h;
		h = CreateEvent(NULL, FALSE, TRUE, _T("IOCTL_LED_STOP_HEARTBEAT"));
		SetEvent(h);

		// NULL the thread handle
		g_hThread = NULL;
	}

	DEBUGMSG(TRUE, (_T("-LED_Close\r\n")));

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Deinit
// 
// Called by the Device Manager to de-initialize a device.
//
extern "C"
__declspec(dllexport)
BOOL LED_Deinit(DWORD hDeviceContext)
{
	DEBUGMSG(TRUE, (_T("+LED_Deinit\r\n")));

	// delete the global critical section
	DeleteCriticalSection(&g_cs);

	// release the alloc'd memory
	UnmapAddress(g_ledAddr);

	// reset the global
	g_hLED = NULL;

	DEBUGMSG(TRUE, (_T("-LED_Deinit\r\n")));

	// return success
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// PowerUp
// 
// stores power to a device.
//
extern "C"
__declspec(dllexport)
void LED_PowerUp(DWORD hDeviceContext)
{
	DEBUGMSG(TRUE, (_T("+LED_PowerUp\r\n")));

	// not implemented, so do nothing.

	DEBUGMSG(TRUE, (_T("-LED_PowerUp\r\n")));

	return;
}

/////////////////////////////////////////////////////////////////////////////
// PowerDown
// 
// Ends power to the device. It is useful only with devices that can be shut 
// off under software control.
//
extern "C"
__declspec(dllexport)
void LED_PowerDown(DWORD hDeviceContext)
{
	DEBUGMSG(TRUE, (_T("+LED_PowerDown\r\n")));

	// not implemented, so do nothing.

	DEBUGMSG(TRUE, (_T("-LED_PowerDown\r\n")));

	return;
}

/////////////////////////////////////////////////////////////////////////////
// DllRegisterServer
//
// This is added for convenience.  Normally DllRegisterServer is exported by COM
// libraries so they can be registered.  We use it to write the driver's registry
// keys and call ActivateDevice to start our driver.
//
// The linker doesn't like this and will generate a LNK4104 warning, but it lets
// us start our driver with the commonly available tool regsvrce.exe rather than
// having to write an installer
//
// Stream interface drivers require a minimal set of registry keys:
// HKEY_LOCAL_MACHINE
//         |
//         +-- Drivers
//                |
//                + BuiltIn
//                     |
//                     + XXX
//                        |
//                        + DLL     string:{dll name}
//                        + Index   dword:{index for series drivers}
//                        + Order   dword:{load order - default to 9}
//                        + Prefix  string:{3-letter "port" moniker}
//
extern "C"
__declspec(dllexport)
BOOL DllRegisterServer(void)
{
	DEBUGMSG(TRUE, (_T("+DllRegisterServer\r\n")));

	HKEY	key		= 0;
	DWORD	ret		= 0;
	DWORD	val		= 0;
	DWORD	disp	= 0;
	HANDLE	hDev	= NULL;
	BYTE	buffer[MAX_PATH];
	
	// create the key
	_stprintf((TCHAR*)buffer, _T("Drivers\\BuiltIn\\%s"), DRIVER_NAME);
	ret = RegCreateKeyEx(HKEY_LOCAL_MACHINE, (TCHAR*)buffer, 0, NULL, 0, 0, NULL, &key, &disp);
	if(ret != ERROR_SUCCESS)
	{
		// failed to create!
		RETAILMSG(TRUE, (_T("  Unable to generate %s driver keys (%i)\r\n"), DRIVER_NAME, ret));
		goto exit;
	}

	// write the 'Dll' value
	_tcscpy((TCHAR*)buffer, DRIVER_LIBRARY);
	ret = RegSetValueEx(key, _T("Dll"), 0, REG_SZ, buffer, (_tcslen(DRIVER_LIBRARY) + 1) * sizeof(TCHAR));
	if(ret != ERROR_SUCCESS)
	{
		// failed to create!
		RETAILMSG(TRUE, (_T("  Unable to generate driver key %s (%i)\r\n"), DRIVER_LIBRARY, ret));
		goto exit;
	}

	// write the 'Prefix' value
	_tcscpy((TCHAR*)buffer, DRIVER_PREFIX);
	ret = RegSetValueEx(key, _T("Prefix"), 0, REG_SZ, buffer, (_tcslen(DRIVER_LIBRARY) + 1) * sizeof(TCHAR));
	if(ret != ERROR_SUCCESS)
	{
		// failed to create!
		RETAILMSG(TRUE, (_T("  Unable to generate driver key %s (%i)\r\n"), DRIVER_PREFIX, ret));
		goto exit;
	}

	DEBUGMSG(TRUE, (_T("-DllRegisterServer\r\n")));

	// write the 'Index' value
	val = DRIVER_INDEX;
	ret = RegSetValueEx(key, _T("Index"), 0, REG_DWORD, (BYTE*)&val, sizeof(DWORD));
	if(ret != ERROR_SUCCESS)
	{
		// failed to create!
		RETAILMSG(TRUE, (_T("  Unable to generate driver key 'Index' (%i)\r\n"), ret));
		goto exit;
	}

	// write the 'Order' value
	val = DRIVER_ORDER;
	ret = RegSetValueEx(key, _T("Order"), 0, REG_DWORD, (BYTE*)&val, sizeof(DWORD));
	if(ret != ERROR_SUCCESS)
	{
		// failed to create!
		RETAILMSG(TRUE, (_T("  Unable to generate driver key 'Order' (%i)\r\n"), ret));
		goto exit;
	}

	// Activate the driver
	_stprintf((TCHAR*)buffer, _T("Drivers\\BuiltIn\\%s"), DRIVER_NAME);
	hDev = ActivateDevice((TCHAR*)buffer, 0);

	if(!hDev)
	{
		ret = GetLastError();
		RETAILMSG(TRUE, (_T("  ActivateDevice failed (%i).\r\n"), ret));
		goto exit;
	}
	else
	{
		DEBUGMSG(TRUE, (_T("  %s loaded.\r\n"), DRIVER_NAME));
	}


	DEBUGMSG(TRUE, (_T("-DllRegisterServer\r\n")));

	return TRUE;

exit:
	if(key)
		RegCloseKey(key);

	return FALSE;

}

/////////////////////////////////////////////
// SetLED
//
// Set the state of the LED(s)
//
void SetLED(DWORD *hLED, int value)
{
	DEBUGMSG(TRUE, (_T("+SetLED : value = %i\r\n"), value));

	// mask the value to make sure we don't write in a "bad" place
	value &= LED_MASK;
	DEBUGMSG(TRUE, (_T("  masked value = %i\r\n"), value));

	DWORD mask = *hLED & ~LED_MASK;

	// set the register value
	*hLED = value | mask;
	DEBUGMSG(TRUE, (_T("-SetLED\r\n")));
}

/////////////////////////////////////////////
// SetLED
//
// Set the state of the LED(s)
//
DWORD WINAPI HeartbeatThreadProc(LPVOID lpParameter)
{
	DEBUGMSG(TRUE, (_T("+HeartbeatThreadProc\r\n")));

	HANDLE hWait = CreateEvent(NULL, FALSE, FALSE, _T("IOCTL_LED_STOP_HEARTBEAT"));

	// run a heartbeat loop
	// note that this will interfere with any other thread trying to use the LED
	while(true)
	{
		if(WaitForSingleObject(hWait, 0) == WAIT_OBJECT_0)
		{
			goto exit;
		}

		SetLED(g_hLED, 1);
		Sleep(750);
		SetLED(g_hLED, 0);
		Sleep(750);
	}

exit:
	ExitThread(0);
	DEBUGMSG(TRUE, (_T("-HeartbeatThreadProc\r\n")));

	return 0;
}

/////////////////////////////////////////////
// MapAddress 
//  Maps the desired memory address.
//  This function should be called first and its
//  return value is then passed to the ReadData16 or
//  WriteData16 functions as the baseAddress.
//
// [I] baseAddress - base address to map
// [O] Virtual Address pointer to mapped region
// 
/////////////////////////////////////////////
LPVOID MapAddress(DWORD baseAddress)
{
	LPVOID pAddr = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
	
	if(pAddr)	
	{
		if(VirtualCopy (pAddr, (PVOID)(baseAddress >> 8), 
				MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
				DEBUGMSG(1, (_T("0x%08x mapped to: 0x%08x\r\n"), baseAddress, pAddr));
		}
		else 
		{
			RETAILMSG(1, (_T("Error! at VirtualCopy()\r\n")));

			return NULL;
		}
	}
	else 
	{
		RETAILMSG(1, (_T("Error! at VirtualAlloc()\r\n")));
		return NULL;
	}

	return pAddr;
}

void UnmapAddress(LPVOID address)
{
	VirtualFree(address, 0, MEM_RELEASE);
	address = NULL;
}

